#pragma once

#include "node.hpp"

#include "data/bonus_points.hpp"
#include "data/csv_dump.hpp"
#include "util/points.hpp"
#include "util/storage_helper.hpp"

#include <QList>
#include <QString>

/**
 * \brief Directly represents the data in the csv-file directly from dynexite in
 * a more accessible manner. Does not change after initialization.
 *
 * Stores data only for students who attended, discards everyone else.
 *
 * \note Data layer 1.
 */
struct DynexiteExam : Node<> {
    using Node<>::Node;

    /**
     * \brief Interprets the CsvDump as a dynexite-csv-file and copies all
     * necessary info. Also runs sanity checks on the given data and throws on
     * error.
     *
     * \exception std::invalid_argument A sanity check on the given csv file
     * failed, see the description in the exception for details, or a std::stod
     * conversion failed.
     * \exception std::out_of_range Thrown by std::stod.
     */
    explicit DynexiteExam(const CsvDump &c) noexcept(false);

    void addBonusPoints(const BonusPoints &bonus) noexcept;

    /**
     * \brief Currently unused.
     */
    QList<QString> mAttemptIds{};
    /**
     * \brief Currently unused. The index for any one student in any of the data
     * structures with an entry for every student is the same, so this is not
     * needed there as a key.
     */
    QList<int> mIdentifier{};

    /**
     * \brief The maximum achievable open, closed and combined points as stated
     * by dynexite. Not influenced by later changes from the user, those are
     * stored in \ref ChangedMaxPoints
     */
    EnumStorage<Points> mPointsMax{};

    /**
     * \brief Open, closed and combined achieved points of each student.
     */
    EnumStorage<QList<Points>> mPoints{};

    /**
     * \brief For each task holds the earned open, closed and combined points
     * each student got in that task, or 0 if the task was not assigned to them.
     */
    EnumStorage<QList<QList<Points>>> mTaskEarnedPoints{};

    /**
     * \brief Holds the maximal (open+closed) points for each task.
     */
    QList<Points> mTaskMaxPoints{};

    // Store the entire csv dump so that we can more easily export it later
    CsvDump mDynexite;
    BonusPoints mBonus;

private:
    // internal helper functions
    void reserveArrays(qsizetype nrStudents, qsizetype nrTasks) noexcept;
    bool insertData(const CsvDump &c,
                    CsvLine::size_type nrTasks) noexcept(false);
};
