#include "yml_parser.hpp"

#include "csv_parser.hpp"
#include "data/data_manager.hpp"
#include "util/enums.hpp"

#include <algorithm>
#include <fstream>
#include <ios>
#include <sstream>
#include <stdexcept>
#include <vector>

namespace impl {

bool toBool(std::string str) {
    std::ranges::transform(str, str.begin(), tolower);
    std::istringstream is(str);
    bool b;
    is >> std::boolalpha >> b;
    return b;
}

} // namespace impl

void writeYml(const std::filesystem::path &p) {

    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    // Open file
    std::ofstream outfile;
    outfile.open(p, std::ios::out | std::ios::trunc);

    // Write to file
    if (outfile.is_open()) {

        outfile << std::boolalpha;

        // YAML identifier
        outfile << "---\n";

        // Version
        outfile << "Version:" << D2R_VERSION << '\n';

        // UserPointConfig
        outfile << "UserPointConfig:" << '\n';
        {
            outfile << "  StepSize:"
                    << as<double>(cur->mUserPointConfig.mGradeBoundaryStepSize)
                    << '\n';
            outfile << "  CountZerothAttempt:"
                    << cur->mUserPointConfig.mCountZerothAttemptAsFirstAttempt
                    << '\n';
        }

        // UserGradeConfig
        outfile << "UserGradeConfig:" << '\n';
        {
            outfile << "  RequirePassAll:"
                    << cur->mUserGradeConfig.mRequirePassAllParts << '\n';
        }

        // ChangedMaxPointsClosed
        outfile << "ChangedMaxPointsClosed:" << '\n';
        {
            outfile << "  ChangedMaxPoints:"
                    << cur->mChangedMaxPointsClosed.mChangedMaxPointBoundary
                           .mMultiple
                    << '\n';
            outfile << "  FollowDefault:"
                    << cur->mChangedMaxPointsClosed.mFollowDefault << '\n';
        }

        // ChangedMaxPointsOpen
        outfile << "ChangedMaxPointsOpen:" << '\n';
        {
            outfile
                << "  ChangedMaxPoints:"
                << cur->mChangedMaxPointsOpen.mChangedMaxPointBoundary.mMultiple
                << '\n';
            outfile << "  FollowDefault:"
                    << cur->mChangedMaxPointsOpen.mFollowDefault << '\n';
        }

        // UserGradeBoundaries
        for (const auto &type : {GradingType::Closed, GradingType::Open}) {
            outfile << "UserGradeBoundaries" << toString(type) << ":" << '\n';
            {
                outfile << "  SetAuto:"
                        << (type == GradingType::Closed
                                ? cur->mClosedUserBoundaries.mSetAutomatically
                                : cur->mOpenUserBoundaries.mSetAutomatically)
                        << '\n';
                for (const auto &g : grades) {
                    outfile << "  Boundary" << static_cast<int>(g) << ":"
                            << cur->getUserGradeBoundaries(type)[g].mMultiple
                            << '\n';
                }
            }
        }

        // Dynexite CSV
        outfile << "Dynexite:" << writeCsv(cur->mExam.mDynexite) << "\\";

        // RWTHonline CSV
        outfile << "RWTHonline:" << writeCsv(cur->mStudents.mRwthOnline)
                << "\\";

        // Bonus CSV
        outfile << "Bonus:" << writeCsv(cur->mExam.mBonus.mBonusPoints) << "\\";
    }
}

struct YamlEntry {
    std::string parent;
    std::string key;
    std::string value;
};

/**
 * \brief Calculate child depth of yml key
 */
int getAndRemoveDepth(std::string &key) {

    const auto strBegin = key.find_first_not_of(' ');
    if (strBegin == std::string::npos)
        return 0;

    key = key.substr(strBegin);

    return static_cast<int>(strBegin) / 2;
}

/**
 * \brief Loads a single yml key-value-pair
 *
 * \param data Where to store the new key-value-pair
 * \param infile Data source
 * \param parent Current parent, might be modified
 * \return Whether or not a new key-value-pair has been read
 */
bool loadYamlKey(std::vector<YamlEntry> &data, std::ifstream &infile,
                 std::string &parent) {
    std::string str;
    // Read key (everything until next colon)
    std::getline(infile, str, ':');

    // Empty key: EOF
    if (str.empty())
        return false;

    YamlEntry newEntry;

    const bool isChild = getAndRemoveDepth(str) > 0;
    if (isChild) {
        newEntry.parent = parent;
    } else {
        newEntry.parent = "";
        // set current key as potential parent for next key
        parent = str;
    }

    newEntry.key = str;

    // Special case for multi-line keys: they end with a "\"
    char endChar = '\n';
    if (str == "Dynexite" || str == "RWTHonline" || str == "Bonus") {
        endChar = '\\';
    }

    std::getline(infile, str, endChar);
    newEntry.value = str;

    data.push_back(newEntry);

    return true;
}

std::unique_ptr<ExamSave>
readStringsToData(const std::vector<YamlEntry> &entries) {
    auto dump = std::make_unique<ExamSave>();

    using namespace impl;

    // set the value for the given parent-key-combination
    for (const auto &[parent, key, value] : entries) {
        if (parent.empty()) {
            if (key == "Dynexite") {
                dump->mDynexite = DynexiteExam(*readCsv(value));
            } else if (key == "RWTHonline") {
                dump->mStudents = StudentData(*readCsv(value));
            } else if (key == "Bonus") {
                dump->mBonus = BonusPoints(*readCsv(value));
            } else if (key == "Version") {
                dump->mSettings.mVersion = value;
            }
        } else if (parent == "UserPointConfig") {
            if (key == "StepSize") {
                dump->mSettings.mUserPointConfigStepSize =
                    toStepSize(std::stod(value));
            } else if (key == "CountZerothAttempt") {
                dump->mSettings.mUserPointConfigCountZero = toBool(value);
            }
        } else if (parent == "UserGradeConfig") {
            if (key == "RequirePassAll") {
                dump->mSettings.mUserGradeConfigRequirePassAll = toBool(value);
            }
        } else if (parent == "ChangedMaxPointsClosed") {
            if (key == "ChangedMaxPoints") {
                dump->mSettings.mChangedMaxPointsClosed = std::stoi(value);
            } else if (key == "FollowDefault") {
                dump->mSettings.mChangedMaxPointsClosedFollowDefault =
                    toBool(value);
            }
        } else if (parent == "ChangedMaxPointsOpen") {
            if (key == "ChangedMaxPoints") {
                dump->mSettings.mChangedMaxPointsOpen = std::stoi(value);

            } else if (key == "FollowDefault") {
                dump->mSettings.mChangedMaxPointsOpenFollowDefault =
                    toBool(value);
            }
        } else if (parent == "UserGradeBoundariesClosed") {
            // To get the grade array index, remove the prefix "Boundary"
            std::string prefix = "Boundary";
            if (key == "SetAuto") {
                dump->mSettings.mUserGradeBoundariesClosedAuto = toBool(value);
            } else if (!key.compare(0, prefix.size(), prefix)) {
                const auto idx = std::stoul(key.substr(prefix.size()));
                dump->mSettings.mUserGradeBoundariesClosed[idx] =
                    std::stoi(value);
            }
        } else if (parent == "UserGradeBoundariesOpen") {
            // To get the grade array index, remove the prefix "Boundary"
            std::string prefix = "Boundary";
            if (key == "SetAuto") {
                dump->mSettings.mUserGradeBoundariesOpenAuto = toBool(value);
            } else if (!key.compare(0, prefix.size(), prefix)) {
                const auto idx = std::stoul(key.substr(prefix.size()));
                dump->mSettings.mUserGradeBoundariesOpen[idx] =
                    std::stoi(value);
            }
        }
    }

    return dump;
}

ExamSave::ExamSave(const CsvDump &dynexite, CsvDump &rwthOnline,
                   const CsvDump &bonus, const YmlSettings &settings)
    : mDynexite(dynexite), mStudents(rwthOnline), mBonus(bonus),
      mSettings(settings) {}

std::unique_ptr<ExamSave> readYml(const std::filesystem::path &p) {

    // Open file
    std::ifstream infile;
    infile.open(p);

    // skip ---
    {
        std::string str;
        std::getline(infile, str, '\n');
        if (str != "---")
            throw std::invalid_argument{"Invalid yaml header"};
    }

    // Read key-by-key until EOF
    std::vector<YamlEntry> readData;
    {
        std::string parent;
        while (loadYamlKey(readData, infile, parent)) {
        }
    }

    // Convert value strings to correct data types
    return readStringsToData(readData);
}
