#include "boundary_window.hpp"

#include "data/data_manager.hpp"

#include <QSignalBlocker>

#include <cassert>

BoundaryWindow::BoundaryWindow(QWidget *parent)
    : QWidget(parent), mLayout(this),
      mRequirePassAllCheckbox(
          "Only pass the exam if all\nexam parts are passed individually"),
      mCountZero("Count \"0 attempts\" as first attempt"),
      mBoundaries{BoundaryTab{GradingType::Closed},
                  BoundaryTab{GradingType::Open}} {

    mStepSize.addItems({"Round grade boundaries: 1",
                        "Round grade boundaries: 0.5",
                        "Round grade boundaries: 0.25"});
    mRounding.addItems({"Round grade boundaries down",
                        "Round grade boundaries up",
                        "Round grade boundaries nearest"});

    mLayout.addWidget(&mRequirePassAllCheckbox);
    mLayout.addWidget(&mCountZero);
    mLayout.addWidget(&mStepSize);
    if constexpr (defaults::showRoundingSetting) {
        mLayout.addWidget(&mRounding);
    }
    mLayout.addWidget(&mTabs);

    setWindowTitle("Options");

    mTabs.addTab(&mBoundaries[GradingType::Open], "Open");
    mTabs.addTab(&mBoundaries[GradingType::Closed], "Closed");

    // set checkbox to correct initial value when points change
    connect(&DataManager::getInstance(), &DataManager::rawDataChanged, this,
            &BoundaryWindow::pointsChanged);
    connect(&DataManager::getInstance(), &DataManager::gradesChanged, this,
            &BoundaryWindow::gradesChanged);

    // receive notification from checkboxes
    connect(&mRequirePassAllCheckbox, &QCheckBox::stateChanged, this,
            &BoundaryWindow::requirePassAllChanged);
    connect(&mCountZero, &QCheckBox::stateChanged, this,
            &BoundaryWindow::countZeroChange);

    // receive notification from combo boxes
    connect(&mStepSize, &QComboBox::activated, this,
            &BoundaryWindow::stepSizeChanged);
    connect(&mRounding, &QComboBox::activated, this,
            &BoundaryWindow::roundingChanged);
}

void BoundaryWindow::pointsChanged() {

    const auto cur = DataManager::getCurrent();

    if (!cur)
        return;

    {
        const bool hasClosed =
            cur->mExam.mPointsMax[GradingType::Closed] > Points{0.};
        const bool hasOpen =
            cur->mExam.mPointsMax[GradingType::Open] > Points{0.};

        mTabs.setTabVisible(0, hasOpen);
        mTabs.setTabVisible(1, hasClosed);
    }
}

void BoundaryWindow::gradesChanged() {

    const auto cur = DataManager::getCurrent();

    if (!cur)
        return;

    {
        QSignalBlocker blocker(&mRequirePassAllCheckbox);

        const auto requirePassAll = cur->mUserGradeConfig.mRequirePassAllParts;

        mRequirePassAllCheckbox.setCheckState(requirePassAll
                                                  ? Qt::CheckState::Checked
                                                  : Qt::CheckState::Unchecked);
    }

    {
        const auto stepSize = cur->mUserPointConfig.mGradeBoundaryStepSize;
        const auto rounding = cur->mUserPointConfig.mGradeBoundaryRounding;
        const auto countZero =
            cur->mUserPointConfig.mCountZerothAttemptAsFirstAttempt;

        QSignalBlocker blocker1(&mStepSize);
        QSignalBlocker blocker2(&mRounding);
        QSignalBlocker blocker3(&mCountZero);

        switch (stepSize) {
        default:
            assert(false);
            [[fallthrough]];
        case GradeBoundaryStepSize::Min0_25:
            mStepSize.setCurrentIndex(2);
            break;
        case GradeBoundaryStepSize::Min0_50:
            mStepSize.setCurrentIndex(1);
            break;
        case GradeBoundaryStepSize::Min1_00:
            mStepSize.setCurrentIndex(0);
            break;
        }

        switch (rounding) {
        default:
            assert(false);
            [[fallthrough]];
        case GradeBoundaryRounding::Down:
            mRounding.setCurrentIndex(0);
            break;
        case GradeBoundaryRounding::Up:
            mRounding.setCurrentIndex(1);
            break;
        case GradeBoundaryRounding::Nearest:
            mRounding.setCurrentIndex(2);
            break;
        }

        if (countZero)
            mCountZero.setCheckState(Qt::Checked);
        else
            mCountZero.setCheckState(Qt::Unchecked);
    }
}

void BoundaryWindow::stepSizeChanged(const int index) {
    const auto cur = DataManager::getCurrent();

    if (!cur)
        return;

    auto currentConfig = cur->mUserPointConfig;
    switch (index) {
    default:
        assert(false);
        [[fallthrough]];
    case 0:
        currentConfig.mGradeBoundaryStepSize = GradeBoundaryStepSize::Min1_00;
        break;
    case 1:
        currentConfig.mGradeBoundaryStepSize = GradeBoundaryStepSize::Min0_50;
        break;
    case 2:
        currentConfig.mGradeBoundaryStepSize = GradeBoundaryStepSize::Min0_25;
        break;
    }

    DataManager::pointConfigChange(currentConfig);
}

void BoundaryWindow::roundingChanged(const int index) {
    const auto cur = DataManager::getCurrent();

    if (!cur)
        return;

    auto currentConfig = cur->mUserPointConfig;
    switch (index) {
    default:
        assert(false);
        [[fallthrough]];
    case 0:
        currentConfig.mGradeBoundaryRounding = GradeBoundaryRounding::Down;
        break;
    case 1:
        currentConfig.mGradeBoundaryRounding = GradeBoundaryRounding::Up;
        break;
    case 2:
        currentConfig.mGradeBoundaryRounding = GradeBoundaryRounding::Nearest;
        break;
    }

    DataManager::pointConfigChange(currentConfig);
}

void BoundaryWindow::requirePassAllChanged(const int state) {
    assert(state == Qt::Unchecked || state == Qt::Checked);

    DataManager::gradeConfigChange(state == Qt::Checked);
}

void BoundaryWindow::countZeroChange(const int state) {
    assert(state == Qt::Unchecked || state == Qt::Checked);

    const auto cur = DataManager::getCurrent();

    if (!cur)
        return;

    auto currentConfig = cur->mUserPointConfig;
    currentConfig.mCountZerothAttemptAsFirstAttempt = state == Qt::Checked;

    DataManager::pointConfigChange(currentConfig);
}
