#include "load_file_view.hpp"

#include "util/storage_helper.hpp"

#include <QFileDialog>
#include <QFrame>
#include <QStyle>

#define HOME_MACHINE 1

QString getFileNameDesc(const FileType t) {
    switch (t) {
    case FileType::dynexite:
        return "Dynexite";
    case FileType::rwth:
        return "RWTHonline";
    case FileType::d2r:
        return "D2R";
    case FileType::bonus:
        return "Bonus points (optional)";
    default:
        assert(false);
        return "Unknown";
    }
}
int getFileHelp(const FileType t, QGridLayout &layout, const QStyle *style,
                QWidget *parent) {

    switch (t) {
    case FileType::dynexite: {
        auto iconLabel = new QLabel("", parent);
        iconLabel->setPixmap(
            style->standardIcon(QStyle::SP_MessageBoxQuestion).pixmap(16));
        layout.addWidget(iconLabel, 0, 0);
        layout.addWidget(new QLabel("CSV export of results", parent), 0, 1);
        layout.addItem(
            new QSpacerItem(0, 0, QSizePolicy::Expanding, QSizePolicy::Minimum),
            0, 2);

        auto line = new QFrame(parent);
        line->setFrameShape(QFrame::HLine);
        layout.addWidget(line, 1, 0, 1, 3);
        return 2;
    }
    case FileType::rwth: {
        auto iconLabel = new QLabel("", parent);
        iconLabel->setPixmap(
            style->standardIcon(QStyle::SP_MessageBoxQuestion).pixmap(16));
        layout.addWidget(iconLabel, 0, 0);
        layout.addWidget(
            new QLabel("CSV export of \"List of Candidates\"", parent), 0, 1);
        layout.addItem(
            new QSpacerItem(0, 0, QSizePolicy::Expanding, QSizePolicy::Minimum),
            0, 2);

        auto line = new QFrame(parent);
        line->setFrameShape(QFrame::HLine);
        layout.addWidget(line, 1, 0, 1, 3);
        return 2;
    }
    case FileType::d2r: {
        auto iconLabel = new QLabel("", parent);
        iconLabel->setPixmap(
            style->standardIcon(QStyle::SP_MessageBoxQuestion).pixmap(16));
        layout.addWidget(iconLabel, 0, 0);
        layout.addWidget(
            new QLabel("Import previously exported d2r-file from this app",
                       parent),
            0, 1);
        layout.addItem(
            new QSpacerItem(0, 0, QSizePolicy::Expanding, QSizePolicy::Maximum),
            0, 2);

        auto line = new QFrame(parent);
        line->setFrameShape(QFrame::HLine);
        layout.addWidget(line, 1, 0, 1, 3);
        return 2;
    }
    case FileType::bonus: {
        auto iconLabel = new QLabel("", parent);
        iconLabel->setPixmap(
            style->standardIcon(QStyle::SP_MessageBoxQuestion).pixmap(16));
        layout.addWidget(iconLabel, 0, 0);
        layout.addWidget(
            new QLabel("CSV export of only \"Bonus points\" from moodle",
                       parent),
            0, 1);
        layout.addItem(
            new QSpacerItem(0, 0, QSizePolicy::Expanding, QSizePolicy::Minimum),
            0, 2);

        auto line = new QFrame(parent);
        line->setFrameShape(QFrame::HLine);
        layout.addWidget(line, 1, 0, 1, 3);
        return 2;
    }
    default:
        // assert(false);
        // return "";
        return 0;
    }
}

LoadFileView::LoadFileView(const FileType type, QWidget *parent)
    : QGroupBox(getFileNameDesc(type), parent), mInputModel(type) {

    // todo: set this, but it also does some annoying stuff - don't do them.
    // mFileNameLabel->setWordWrap(true);

    connect(&mBrowseButton, &QPushButton::clicked, this,
            &LoadFileView::handleBrowseButton);

    // First, an explanatory text
    const auto nextRow = getFileHelp(type, mOuterLayout, style(), this);
    mOuterLayout.addLayout(&mHLayout, nextRow, 0, 1, 3);

    // the elements of each file selection are arranged horizontally within
    // the box
    mHLayout.addWidget(&mFileNameLabel);
    mHLayout.addWidget(&mFileParseStatusIcon);
    mHLayout.addWidget(&mBrowseButton);
    setLayout(&mHLayout);
}

void LoadFileView::handleBrowseButton() {
    // window captions for the file selection dialog for each file type
    static constexpr EnumStorage<const char *, FileType> captionTexts = {
        "Open exam results file from dynexite",
        "Open student info file from RWTHonline",
        "Open eTest bonus points file", "Open D2R file"};

    static constexpr auto csvFilter =
        "Comma-separated values (CSV) file (*.csv);;All files (*.*)";
    static constexpr auto d2rFilter =
        "Configured exam (D2R) file (*.d2r);;All files (*.*)";

    // let the user select a file
    const auto fileName = QFileDialog::getOpenFileName(
        this, captionTexts[mInputModel.mType], "",
        mInputModel.mType == FileType::d2r ? d2rFilter : csvFilter);

    // if the user didn't select any files, don't change anything
    if (!fileName.isEmpty()) {
        // write the path to the selected file to the label of the corresponding
        // file type
        mFileNameLabel.setText(fileName);

        if (mInputModel.loadCsv(mFileNameLabel.text().toStdWString())) {
            setFileStatus(FileSelectState::ok);
        } else {
            setFileStatus(FileSelectState::error);
        }
    }
}

void LoadFileView::reset() {
    // remove input from the labels displaying the selected files, reset icons,
    // unload csv-file
    mFileNameLabel.setText("");

    // We deliberately keep the current data loaded in the model, so that we can
    // reset the view and extract the data from the model (which clears it
    // there)

    setFileStatus(FileSelectState::empty);
}

void LoadFileView::tempSetDeveloperPaths() {
#if defined(DYNEXITE_EXAMPLE_FILE_PATH) && defined(RWTHONLINE_EXAMPLE_FILE_PATH)
    if (mInputModel.mType == FileType::dynexite) {
        mFileNameLabel.setText(DYNEXITE_EXAMPLE_FILE_PATH);
    } else {
        mFileNameLabel.setText(RWTHONLINE_EXAMPLE_FILE_PATH);
    }

    if (mInputModel.loadCsv(mFileNameLabel.text().toStdWString())) {
        setFileStatus(FileSelectState::ok);
    } else {
        setFileStatus(FileSelectState::error);
    }
#endif
}

void LoadFileView::setFileStatus(const FileSelectState status) {
    // icons corresponding to each file select state
    static constexpr EnumStorage<QStyle::StandardPixmap, FileSelectState>
        icons = {
            QStyle::SP_MessageBoxQuestion, // empty, no icon, not used
            QStyle::SP_MessageBoxCritical,
            QStyle::SP_DialogOkButton // todo: find better icon
        };

    // update status
    mFileState = status;

    // update icon
    if (status != FileSelectState::empty) {
        mFileParseStatusIcon.setPixmap(
            style()->standardIcon(icons[status]).pixmap(16));
    } else {
        mFileParseStatusIcon.setText(" ");
    }
    emit modelChange();
}
